/*
* Copyright (C) 2009, Oy L M Ericsson Ab, NomadicLab <pubsub@nomadiclab.com>
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License version 2 as
* published by the Free Software Foundation.
*
* Alternatively, this software may be distributed under the terms of the BSD
* license.
*
* See LICENSE and COPYING for more details.
*/

#define SOCKET_IPC

/* kqueue/kevent */
#include <sys/types.h>
#include <sys/event.h>
#include <sys/time.h>

/* psirpd_net.h */
#include <sys/socket.h>
#include <net/if_dl.h>
#include <net/if.h>
#include <sys/ioctl.h>

#include <netinet/in.h> /* XXX */
#include <sys/uio.h>

#include <string.h>
#include <errno.h>
#include <unistd.h>

/* libpsirp */
#include <sys/param.h>
#include <libpsirp.h>
#include "../../libpsirp/src/psirp_debug.h"

#include "psirp_common_types.h"
#include "psirp.h"
#include "psirpd_hdrs.h"
#include "psirpd_ipc.h"

#include "psirpd_net.h"
#include "psirpd_out_q.h"
#include "psirpd_packet.h"


static psirp_id_t ipc_net_rid;
static psirp_id_t ipc_net_sid;
#if 0
static psirp_pub_t ipc_net_pub;
#endif

static psirp_id_t ipc_local_rid;
static psirp_id_t ipc_local_sid;
static psirp_pub_t ipc_local_pub;

static psirp_pub_t active_pub; /* XXXX */
static psirpd_hdrs_metadata_t active_pub_md; /* XXXX */

#ifdef SOCKET_IPC /* XXX */
static int ipc_net_sock;
static struct sockaddr_in ipc_net_addr;

static int ipc_local_sock;
static struct sockaddr_in ipc_local_addr;
#endif


static psirp_error_t psirpd_ipc_chevents(int, u_int, int*);
static psirp_error_t psirpd_ipc_createpub(psirp_id_t *, psirp_id_t *,
                                          psirp_pub_t *,
					  void *, int,
                                          int);


psirp_error_t
psirpd_ipc_init(const char *ipc_local_rid_str, const char *ipc_net_rid_str) {
    
    psirp_pub_t ipc_net_pub = NULL;
    int err = 0;
    
#ifdef SOCKET_IPC /* XXX */
    ENTER();

    int *socks[] = { &ipc_net_sock, &ipc_local_sock };
    struct sockaddr_in *addrs[] = { &ipc_net_addr, &ipc_local_addr };
    int i;
    
    for (i = 0; i < 2; i++) {
	int *sockp = socks[i];
	struct sockaddr_in *addrp = addrs[i];
	int r;
	
	*sockp = -1;
	memset(addrp, 0, sizeof(*addrp));
	
	*sockp = socket(AF_INET, SOCK_DGRAM, 0);
	PSIRP_ETL(0 > *sockp, PSIRP_FAIL, PSIRP_DBG_ERR,
		  "socket(): [%d] %s", errno, strerror(errno));
	
	addrp->sin_family      = AF_INET;
	addrp->sin_addr.s_addr = htonl(INADDR_LOOPBACK);
	if (i == 0) {
	    addrp->sin_port    = htons(0xacdc);
	}
	else {
	    addrp->sin_port    = htons(0xdcac);
	    
	    r = bind(*sockp, (struct sockaddr *)addrp, sizeof(*addrp));
	    PSIRP_ETL(0 > r,  PSIRP_FAIL, PSIRP_DBG_ERR,
		      "bind(): [%d] %s", errno, strerror(errno));
	}
    }
    
    ipc_local_pub = NULL;
    
    RETURN(PSIRP_OK);
#endif
    ENTER();
    
    PSIRP_ETL(NULL == ipc_local_rid_str, PSIRP_FAIL_NULL_POINTER,
              PSIRP_DBG_ERR, "No IPC RId string for local events");
    PSIRP_ETL(NULL == ipc_net_rid_str, PSIRP_FAIL_NULL_POINTER,
              PSIRP_DBG_ERR, "No IPC RId string for network events");
    /* + Compare? */
    PSIRP_DEBUG(PSIRP_DBG_GARB, "ipc_local_rid = %s" , ipc_local_rid_str);
    PSIRP_DEBUG(PSIRP_DBG_GARB, "ipc_net_rid = %s" , ipc_net_rid_str);
    
    /* Subscribe to the IPC SId/RId for local events */
    PSIRP_DEBUG(PSIRP_DBG_GARB,
                "Subscribing to IPC publication for local events");
    
    memset(&ipc_local_sid, 0, sizeof(ipc_local_sid)); /* XXX: Scope 0 */
    ipc_local_sid.id[PSIRP_ID_LEN-2] = 0xbb;      /* XXX: bb00 */
    
    memset(&ipc_local_rid, 0, sizeof(ipc_local_rid));
    PSIRP_ETL(psirp_atoid(&ipc_local_rid, ipc_local_rid_str),
	      PSIRP_FAIL_INVALIDARG,
              PSIRP_DBG_ERR, "psirp_atoid()");
    
    ipc_local_pub = NULL;
    err = psirp_subscribe(&ipc_local_sid, &ipc_local_rid, &ipc_local_pub);
    
    if (0 != err && (ENOENT == errno || /* XXX: */ ESRCH == errno)) {
        psirp_pub_t tmp_pub = NULL;
        
        PSIRP_DEBUG(PSIRP_DBG_WARN,
                    "psirp_subscribe(): [%d] %s",
                    errno, strerror(errno));
        
        /*
         * An IPC publication did not exist, so create and publish a
         * dummy one in order to be able to subscribe.
         */
        PSIRP_DEBUG(PSIRP_DBG_INFO,
                    "No IPC publication exists, creating a dummy one");
        PSIRP_EFL(psirpd_ipc_createpub(&ipc_local_sid, &ipc_local_rid,
                                       &tmp_pub,
				       NULL, sizeof(psirpd_hdrs_metadata_t),
                                       PSIRP_TRUE),
                  PSIRP_DBG_ERR);
        psirp_free(tmp_pub);
        tmp_pub = NULL;
        
	/* Wait for a short moment. */
	sleep(1);
	
        /* Try subscribing again. */
        err = psirp_subscribe(&ipc_local_sid, &ipc_local_rid, &ipc_local_pub);
    }
    
    PSIRP_ETL(0 != err, PSIRP_FAIL, PSIRP_DBG_ERR,
              "psirp_subscribe(): [%d] %s", errno, strerror(errno));
    
    /* Create a publication for the IPC SId/RId for network events. */
    PSIRP_DEBUG(PSIRP_DBG_GARB,
                "Creating IPC publication for network events");

    memset(&ipc_net_sid, 0, sizeof(ipc_net_sid)); /* XXX: Scope 0 */
    ipc_net_sid.id[PSIRP_ID_LEN-2] = 0xbb;        /* XXX: ::bb00 */
    
    memset(&ipc_net_rid, 0, sizeof(ipc_net_rid));
    PSIRP_ETL(psirp_atoid(&ipc_net_rid, ipc_net_rid_str),
	      PSIRP_FAIL_INVALIDARG,
              PSIRP_DBG_ERR, "psirp_atoid()");
    
    ipc_net_pub = NULL;
    PSIRP_EFL(psirpd_ipc_createpub(&ipc_net_sid, &ipc_net_rid,
                                   &ipc_net_pub,
				   NULL, sizeof(psirpd_hdrs_metadata_t),
                                   PSIRP_TRUE),
              PSIRP_DBG_ERR);
    psirp_free(ipc_net_pub);
    
    /* XXXX */
    active_pub = NULL;
    memset(&active_pub_md, 0, sizeof(active_pub_md));
    
    RETURN(PSIRP_OK);
}

static psirp_error_t
psirpd_ipc_createpub(psirp_id_t *sidp, psirp_id_t *ridp,
		     psirp_pub_t *pubp,
		     void *data, int len,
		     int publish) {
    ENTER();
    
    *pubp = NULL;
    PSIRP_ETLB(0 != psirp_create(len, pubp),
	       PSIRP_FAIL,
	       PSIRP_DBG_ERR,
	       {
		   *pubp = NULL;
	       },
	       "psirp_create(): [%d] %s", errno, strerror(errno));
    PSIRP_ETLB(NULL == psirp_pub_data(*pubp),
               PSIRP_FAIL,
               PSIRP_DBG_ERR,
               {
                   psirp_free(*pubp);
                   *pubp = NULL;
               },
               "No publication data");
    if (NULL != data)
	memcpy(psirp_pub_data(*pubp), data, len);
    else
	memset(psirp_pub_data(*pubp), 0, psirp_pub_data_len(*pubp));

    if (PSIRP_TRUE == publish) {
        PSIRP_ETLB(0 != psirp_publish(sidp, ridp, *pubp),
                   PSIRP_FAIL,
                   PSIRP_DBG_ERR,
                   {
                       psirp_free(*pubp);
                       *pubp = NULL;
                   },
                   "psirp_publish(): [%d] %s", errno, strerror(errno));
    }
    
    RETURN(PSIRP_OK);
}

void
psirpd_ipc_cleanup()
{
    ENTER();

    //psirpd_ipc_local_unregevents(kq); /* XXX: ? */
    
    /* XXX: Unmap stuff etc.? */

#ifdef SOCKET_IPC /* XXX */
    close(ipc_net_sock);
    close(ipc_local_sock);
#endif
    
    if (NULL != ipc_local_pub) {
	psirp_free(ipc_local_pub);
    }
    
    RETURN;
}

static psirp_error_t
psirpd_ipc_chevents(int kq, u_int flags, int *fd) {
    
    static struct kevent cl[1];
    
    ENTER();

#ifdef SOCKET_IPC /* XXX */
    *fd = ipc_local_sock;
    
    EV_SET(&cl[0],      /* kevent, */
	   *fd,         /* ident,  */
	   EVFILT_READ, /* filter, */
           flags,       /* flags,  */
	   0, 0, 0);    /* fflags, data, udata */
#else
    *fd = psirp_pub_fd(ipc_local_pub);
    
    EV_SET(&cl[0],                    /* kevent, */
	   *fd,                       /* ident,  */
	   EVFILT_VNODE,              /* filter, */
           flags,                     /* flags,  */
	   NOTE_PUBLISH /* | NOTE_UNMAP */, /* fflags, */ /* XXX: ? */
	   0,                         /* data,   */
           (void *)ipc_local_pub);    /* udata   */
#endif
    PSIRP_ETL(0 > kevent(kq,      /* kqueue,                */
                         cl, 1,   /* *changelist, nchanges, */
                         NULL, 0, /* *eventlist, nevents,   */
                         NULL),   /* *timeout               */
              PSIRP_FAIL,
              PSIRP_DBG_ERR,
	      "kevent(): [%d] %s", errno, strerror(errno));
    
    RETURN(PSIRP_OK);
}

psirp_error_t
psirpd_ipc_local_regevents(int kq, int *fd) {
    return psirpd_ipc_chevents(kq, EV_ADD | EV_CLEAR, fd);
}

psirp_error_t
psirpd_ipc_local_unregevents(int kq) {
    int dummy_fd;
    return psirpd_ipc_chevents(kq, EV_DELETE, &dummy_fd);
}

psirp_error_t
psirpd_ipc_local_handle_event(struct kevent *el) {
    
    psirpd_hdrs_metadata_t *ipc_local_pub_data = NULL;
    
#ifdef SOCKET_IPC /* XXX */
    psirpd_ipc_hdrs_metadata_ext_t md_hdr_ext;
    int r;
    
    ENTER();
    
    memset(&md_hdr_ext, 0, sizeof(md_hdr_ext));
    
    r = read(ipc_local_sock, &md_hdr_ext, sizeof(md_hdr_ext));
    PSIRP_ETL(0 > r || sizeof(psirpd_hdrs_metadata_t) > r,
	      PSIRP_FAIL, PSIRP_DBG_ERR,
	      "read(): [%d] %s\n", errno, strerror(errno));
    
    ipc_local_pub_data = (psirpd_hdrs_metadata_t *)&md_hdr_ext;
#else
    ENTER();
    
    PSIRP_ETL(0 != psirp_subscribe(&ipc_local_sid, &ipc_local_rid,
				   &ipc_local_pub),
	      PSIRP_FAIL, PSIRP_DBG_ERR,
	      "psirp_subscribe(): [%d] %s", errno, strerror(errno));
    /* XXX: Unmap old publication? */
    
    PSIRP_ET(psirp_pub_data_len(ipc_local_pub) < sizeof(psirpd_hdrs_metadata_t),
             PSIRP_FAIL,
             "Not an IPC publication");
    
    ipc_local_pub_data =
	(psirpd_hdrs_metadata_t *)psirp_pub_data(ipc_local_pub);
#endif
        
    switch (ipc_local_pub_data->rzvhdr_type) {
    case PSIRP_HDR_RZV_PUBLISH_METADATA:
	PSIRP_DEBUG(PSIRP_DBG_GARB, "Event: publish metadata");
	PSIRP_EF(psirpd_packet_out_metadata(&ipc_local_pub_data->sid,
					    &ipc_local_pub_data->rid,
					    &ipc_local_pub_data->fid,
					    ipc_local_pub_data->len));
	break;
    case PSIRP_HDR_RZV_PUBLISH_DATACHUNK:
	PSIRP_DEBUG(PSIRP_DBG_GARB, "Event: publish datachunk");
	psirpd_packet_out_queuechunks(&ipc_local_pub_data->sid,
				      &ipc_local_pub_data->rid,
				      &ipc_local_pub_data->fid);
        break;
    case PSIRP_HDR_RZV_SUBSCRIBE_METADATA:
	PSIRP_DEBUG(PSIRP_DBG_GARB, "Event: subscribe metadata");
	PSIRP_EF(psirpd_packet_out_submetadata(&ipc_local_pub_data->sid,
					       &ipc_local_pub_data->rid,
					       &ipc_local_pub_data->fid));
        break;
    case PSIRP_HDR_RZV_SUBSCRIBE_DATACHUNK:
	PSIRP_DEBUG(PSIRP_DBG_INFO, "Event: subscribe datachunk");
	{
	    psirpd_ipc_metadata_ext_t *md_ext;
	    psirp_fid_t *md_fid;
	    
#ifdef SOCKET_IPC /* XXX */
	    PSIRP_ETL(sizeof(psirpd_ipc_hdrs_metadata_ext_t) > r,
		      PSIRP_FAIL, PSIRP_DBG_ERR,
		      "Expected extended metadata");
	    md_ext = &md_hdr_ext.md_ext;
	    
#else
	    PSIRP_ETL(psirp_pub_data_len(ipc_local_pub)
		      < sizeof(psirpd_hdrs_metadata_ext_t),
		      PSIRP_FAIL, PSIRP_DBG_ERR,
		      "Expected extended metadata");
	    md_ext =
		&((psirpd_ipc_hdrs_metadata_ext_t *)
		  psirp_pub_data(ipc_local_pub))->md_ext;
#endif
	    md_fid = &md_ext->fid;
	    
	    /* XXXX */
	    if (NULL == active_pub && !md_ext->relay) {
		PSIRP_DEBUG(PSIRP_DBG_INFO, "Set active publication metadata, SId/RId: %s, len: %d", psirp_idstoa(&ipc_local_pub_data->sid, &ipc_local_pub_data->rid), ipc_local_pub_data->len);
		memcpy(&active_pub_md.sid, &ipc_local_pub_data->sid,
		       sizeof(psirp_id_t));
		memcpy(&active_pub_md.rid, &ipc_local_pub_data->rid,
		       sizeof(psirp_id_t));
		active_pub_md.len = ipc_local_pub_data->len;
		
		PSIRP_DEBUG(PSIRP_DBG_GARB, "Create active publication");
		PSIRP_EFL(psirpd_ipc_createpub(NULL, NULL,
					       &active_pub,
					       NULL, active_pub_md.len,
					       PSIRP_FALSE),
			  PSIRP_DBG_ERR);
	    }
	    
	    { /* For debugging: */
		char fid1[PSIRP_ID_LEN*2+1];
		char fid2[PSIRP_ID_LEN*2+1];
		
		memset(fid1, 0, sizeof(fid1));
		memset(fid2, 0, sizeof(fid2));
		
		memcpy(fid1, psirp_idtoa(&ipc_local_pub_data->fid),
		       sizeof(psirp_fid_t));
		memcpy(fid2, psirp_idtoa(md_fid),
		       sizeof(psirp_fid_t));
		
		PSIRP_DEBUG(PSIRP_DBG_INFO,
			    "Subscription metadata:\n"
			    "         SId/RId: %s\n"
			    "         FId:     %s\n"
			    "         MD-FId:  %s",
			    psirp_idstoa(&ipc_local_pub_data->sid,
					 &ipc_local_pub_data->rid),
			    fid1, fid2);
	    }
	    
	    psirpd_packet_out_subdata(&ipc_local_pub_data->sid,
				      &ipc_local_pub_data->rid,
				      &ipc_local_pub_data->fid,
				      md_fid,
				      md_ext->relay);
	}
        break;
    default:
        PSIRP_ET(PSIRP_TRUE, PSIRP_FAIL, "Unknown IPC publication type");/*XXX*/
        break;
    }
    
    RETURN(PSIRP_OK);
}

psirp_error_t
psirpd_ipc_net_handle_metadata(psirpd_hdrs_metadata_t *md_hdr) {
    /*
     * We have received publication metadata or a subscription from
     * the network. We pass it on to the local rendezvous daemon by
     * creating and publishing an IPC publication.
     */
    
    psirp_pub_t ipc_net_pub = NULL;
    
    ENTER();
    
    PSIRP_DEBUG(PSIRP_DBG_GARB,
		"hdr_type=0x%02x, nxt_hdr=0x%02x, rzvhdr_type=0x%02x, "
		"sid/rid=%s, ...",
		md_hdr->hdr_type, md_hdr->nxt_hdr, md_hdr->rzvhdr_type,
		psirp_idstoa(&md_hdr->sid, &md_hdr->rid));
    
    PSIRP_DEBUG(PSIRP_DBG_GARB, "Publish metadata as IPC publication");
#ifdef SOCKET_IPC /* XXX */
    ipc_net_pub = NULL;
    PSIRP_ETL(0 > sendto(ipc_net_sock,
			 md_hdr, sizeof(*md_hdr),
			 0,
			 (struct sockaddr *)&ipc_net_addr,
			 sizeof(ipc_net_addr)),
	      PSIRP_FAIL, PSIRP_DBG_ERR,
	      "sendto(): [%d] %s\n", errno, strerror(errno));
#else
    PSIRP_EFL(psirpd_ipc_createpub(&ipc_net_sid, &ipc_net_rid,
                                   &ipc_net_pub,
				   md_hdr, sizeof(*md_hdr),
                                   PSIRP_TRUE),
              PSIRP_DBG_ERR);
    psirp_free(ipc_net_pub);
#endif
    
    RETURN(PSIRP_OK);
}

psirp_error_t
psirpd_ipc_net_handle_data(psirpd_hdrs_rzvhdr_t *rzvhdr,
			   void *payload, u_int64_t len) {
    
    ENTER();

    PSIRP_DEBUG(PSIRP_DBG_GARB,
		"active_pub=%p, active_pub_md.sid/rid=%s, seqnum=%lu",
		active_pub, psirp_idstoa(&active_pub_md.sid,
					 &active_pub_md.rid),
		rzvhdr->seqnum);
    
    /* XXXX */
    if (rzvhdr->hdr_type == PSIRP_HDR_RZV_PUBLISH_DATACHUNK
	&& NULL != active_pub
	&& !memcmp(&rzvhdr->sid, &active_pub_md.sid, sizeof(psirp_id_t))
	&& !memcmp(&rzvhdr->rid, &active_pub_md.rid, sizeof(psirp_id_t))) {
	
	PSIRP_DEBUG(PSIRP_DBG_GARB,
		    "Copying data chunk to publication, "
		    "seqnum=%d", rzvhdr->seqnum);
	memcpy(psirp_pub_data(active_pub)+PSIRP_CHUNK_SIZE*rzvhdr->seqnum,
	       payload,
	       len);
	
	if (active_pub_md.max_seqnum > 0
	    && rzvhdr->seqnum != active_pub_md.max_seqnum-1) {
	    
	    PSIRP_DEBUG(PSIRP_DBG_WARN,
			"Seqnum: %d != %d-1, packet lost?",
			rzvhdr->seqnum, active_pub_md.max_seqnum);
	}
	active_pub_md.max_seqnum = rzvhdr->seqnum;
	
	if (rzvhdr->seqnum == 0) {
	    PSIRP_DEBUG(PSIRP_DBG_INFO, "Last publication data chunk received");
	    PSIRP_ETLB(0 != psirp_publish(&active_pub_md.sid,
					  &active_pub_md.rid,
					  active_pub),
		       PSIRP_FAIL,
		       PSIRP_DBG_ERR,
		       {
			   psirp_free(active_pub);
			   active_pub = NULL;
			   memset(&active_pub_md, 0, sizeof(active_pub_md));
		       },
		       "psirp_publish(): [%d] %s", errno, strerror(errno));
	    
	    psirp_free(active_pub);
	    active_pub = NULL;
	    memset(&active_pub_md, 0, sizeof(active_pub_md));
	}
    }
    else {
	PSIRP_ETL(PSIRP_TRUE, PSIRP_FAIL, PSIRP_DBG_ERR,
		  "Received data chunk for non-active publication, "
		  "SId/RId: %s", psirp_idstoa(&rzvhdr->sid, &rzvhdr->rid));
    }
    
    RETURN(PSIRP_OK);
}
