/*
* Copyright (C) 2009, Oy L M Ericsson Ab, NomadicLab <pubsub@nomadiclab.com>
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License version 2 as
* published by the Free Software Foundation.
*
* Alternatively, this software may be distributed under the terms of the BSD
* license.
*
* See LICENSE and COPYING for more details.
*/

//#include <pcap.h>
//#include <libnet11/libnet.h>

/* psirpd_net.h */
#include <sys/types.h>
#include <sys/socket.h>
#include <net/if_dl.h>
#include <net/if.h>

#include <string.h>
#include <stdlib.h>

#ifdef PSIRP_PLA
#include "libpla.h"
#endif

/* libpsirp */
#include <sys/param.h>
#include <libpsirp.h>

#include "psirp_common_types.h"
#include "psirpd_types.h"
#include "psirp.h"
//#include "psirp_debug.h"
//#include "psirp_io.h"
//#include "psirpd_pl.h"
#include "psirpd_hdrs.h"
#include "psirpd_net.h"
#include "psirpd_out_q.h"
#include "psirpd_packet.h"
#include "psirpd_rzv.h"
#include "psirpd_sec_pla.h"
#include "psirpd_ipc.h"

#include "../../libpsirp/src/psirp_debug.h"


static u_int8_t psirpd_rzv_rid_policy = PSIRP_RZV_RID_RANDOM; 


psirp_error_t
psirpd_rzv_init(u_int8_t rid_policy) {

    ENTER();

    switch(rid_policy) {
    case PSIRP_RZV_RID_RANDOM:
    case PSIRP_RZV_RID_PLA:
        psirpd_rzv_rid_policy = rid_policy;
        break;
    default:
        PSIRP_DEBUG(PSIRP_DBG_ERR, "Unnkown RID policy type: %d", 
                    rid_policy);       
        RETURN PSIRP_FAIL_UNKNOWN_RID_POLICY;
    }

    RETURN PSIRP_OK;
}

void
psirpd_rzv_cleanup(void)
{
    ENTER();

    RETURN;
}


psirp_error_t
psirpd_rzv_create_rid(psirp_id_t *rid) {

    int i = 0;

    ENTER();

    switch(psirpd_rzv_rid_policy) {
    case PSIRP_RZV_RID_RANDOM:
        memset(rid, 0,  sizeof(psirp_id_t));
        for (i = 0; i < PSIRP_ID_LEN; i++) {
            rid->id[i] = (u_int8_t)random();
        }
        break;
    case PSIRP_RZV_RID_PLA:
        PSIRP_EF(psirpd_sec_pla_get_rid(rid));
        break;
    default:
        PSIRP_DEBUG(PSIRP_DBG_ERR, "Unknown RID policy type: %d", 
                    psirpd_rzv_rid_policy);       
        RETURN PSIRP_FAIL_UNKNOWN_RID_POLICY;
    }

    RETURN PSIRP_OK;
}






/** 
 * Handler that is called for incoming data chunk packet
 * @param pkt_ctx pointer to the packet context
 * @return On sucess: PSIRP_OK <br>On failure: PSIRP_FAIL_ error code see 
 *   psirpd_types.h
 */
psirp_error_t 
psirpd_rzv_rcv_datachunk(pkt_ctx_t *pkt_ctx) {

    psirpd_hdrs_rzvhdr_t  *rzv_hdr = NULL;

    ENTER();

    PSIRP_ETL(pkt_ctx == NULL, PSIRP_FAIL_NULL_POINTER, 
              PSIRP_DBG_ERR, "pkt_ctx == NULL");
    
    rzv_hdr = pkt_ctx->rzv_hdr;
    PSIRP_ETL(rzv_hdr == NULL, PSIRP_FAIL_NULL_POINTER, 
              PSIRP_DBG_ERR, "pkt_ctx->rzv_hdr == NULL");

#if 0
    PSIRP_EF(psirpd_pl_store_data((u_int8_t*)&rzv_hdr->sid,
                                  (u_int8_t*)&rzv_hdr->rid,
                                  rzv_hdr->seqnum,
                                  0, 
                                  pkt_ctx->payload,
                                  NULL));
#endif

    psirpd_ipc_net_handle_data(rzv_hdr, pkt_ctx->payload,
			       pkt_ctx->offset-pkt_ctx->payload); /* XXX: ? */
    
    RETURN PSIRP_OK;
}

/** 
 * Handler that is called for incoming metadata packet
 * @param pkt_ctx pointer to the packet context
 * @return On sucess: PSIRP_OK <br>On failure: PSIRP_FAIL_ error code see 
 *   psirpd_types.h
 */
psirp_error_t 
psirpd_rzv_rcv_metadata(pkt_ctx_t *pkt_ctx) {
    
    psirpd_hdrs_metadata_t *md_hdr = NULL;
    //psirpd_hdrs_rzvhdr_t  *rzv_hdr = NULL;

    ENTER();

    PSIRP_ETL(pkt_ctx == NULL, PSIRP_FAIL_NULL_POINTER, 
              PSIRP_DBG_ERR, "pkt_ctx == NULL");
    
    md_hdr = pkt_ctx->md_hdr;
    PSIRP_ETL(md_hdr == NULL, PSIRP_FAIL_NULL_POINTER, 
              PSIRP_DBG_ERR, "pkt_ctx->md_hdr == NULL");

#if 0
    rzv_hdr = pkt_ctx->rzv_hdr;
    PSIRP_ETL(rzv_hdr == NULL, PSIRP_FAIL_NULL_POINTER, 
              PSIRP_DBG_ERR, "pkt_ctx->rzv_hdr == NULL");
    /* We should not ignore the RZV header. For example, we could take
     * the SId and RId and publish the rest of the packet with those
     * Ids and expect that some entity will process the packet
     * further. */
#endif

    PSIRP_DEBUG(PSIRP_DBG_GARB, "md_hdr->len = %lu", md_hdr->len);

    PSIRP_EF(psirpd_ipc_net_handle_metadata(md_hdr));

    RETURN PSIRP_OK;
}

/** 
 * Handler that is called for incoming data subscription packet
 * @param pkt_ctx pointer to the packet context
 * @return On sucess: PSIRP_OK <br>On failure: PSIRP_FAIL_ error code see 
 *   psirpd_types.h
 */
psirp_error_t 
psirpd_rzv_rcv_subscribe_data(pkt_ctx_t *pkt_ctx) {
    
    psirpd_hdrs_metadata_t *md_hdr = NULL;
    //out_q_list_item_t *out_q_item = NULL;
    //psirpd_hdrs_bitmap_t *bmap_hdr = NULL;
    //uint8_t *data_map = NULL;
    //int size;

#if 0
    psirpd_hdrs_rzvhdr_t  *rzv_hdr = NULL;
#endif

    ENTER();

    PSIRP_ETL(pkt_ctx == NULL, PSIRP_FAIL_NULL_POINTER, 
              PSIRP_DBG_ERR, "pkt_ctx == NULL");

    md_hdr = pkt_ctx->md_hdr;
    PSIRP_ETL(md_hdr == NULL, PSIRP_FAIL_NULL_POINTER, 
              PSIRP_DBG_ERR, "pkt_ctx->md_hdr == NULL");

#if 0
    rzv_hdr = pkt_ctx->rzv_hdr;
    PSIRP_ETL(rzv_hdr == NULL, PSIRP_FAIL_NULL_POINTER, 
              PSIRP_DBG_ERR, "pkt_ctx->rzv_hdr == NULL");
#endif

#ifdef PSIRP_BMAP
    PSIRP_EF(psirpd_pl_find((u_int8_t*)&md_hdr->sid,
                            (u_int8_t*)&md_hdr->rid,
                            &pl));

    PSIRP_ETL(pl == NULL, PSIRP_FAIL_NULL_POINTER, 
              PSIRP_DBG_ERR, "pl == NULL");

    bmap_hdr = pkt_ctx->bmap_hdr;
    PSIRP_ETL(bmap_hdr == NULL, PSIRP_FAIL_NULL_POINTER, 
              PSIRP_DBG_ERR, "pkt_ctx->bmap_hdr == NULL");

    
    /* TODO: If we do not have a publication that
     * matches with the received rid we should
     * store the received subscription. After the corresponding
     * publication is created/cached we can match it with 
     * the stored subscription. */

    /* TODO: We should check if fid is zero. 
     * Pass the subscriber ID (i.e. md_hdr->subscriber_ID)
     * to another module to form the fid */

    if (bmap_hdr->compression == PSIRP_COMP_NOTUSED)
        data_map = bmap_hdr->bitmap;
    else if (bmap_hdr->compression == PSIRP_COMP_RLE) {
	size = ntohs(bmap_hdr->bitmap_len);
        data_map = psirpd_bmap_rle_decode(bmap_hdr->bitmap, 
                                          &size, PSIRP_DATAMAP_SIZE);
    }

    if (!data_map) {
        PSIRP_DEBUG(PSIRP_DBG_ERR, "Bitmap not received");
        RETURN PSIRP_FAIL_INVALIDARG;
    }

    out_q_item = psirpd_out_q_create_item(pl,
                                          &md_hdr->fid,
                                          &md_hdr->sid,
                                          &md_hdr->rid,
                                          data_map);

    if (bmap_hdr->compression == PSIRP_COMP_RLE) {
        PSIRP_FREE(data_map);
    }

    PSIRP_ETL(out_q_item == NULL, PSIRP_FAIL_NULL_POINTER, 
              PSIRP_DBG_ERR, "out_q_item == NULL");
#if 0        
    PSIRP_EF(psirpd_bmap_import(out_q_item->data_map, 
                                md_hdr->data_map,
                                PSIRP_DATAMAP_SIZE));

    psirpd_bmap_print_map(out_q_item->data_map, 0);
#endif
    psirpd_out_q_add_item(out_q_item);
    
    PSIRP_EF(psirpd_out_q_send());
#endif /* PSIRP_BMAP */

    psirpd_ipc_net_handle_metadata(md_hdr);
    
    RETURN PSIRP_OK;
}

/** 
 * Handler that is called for incoming metadata subscription packet
 * @param pkt_ctx pointer to the packet context
 * @return On sucess: PSIRP_OK <br>On failure: PSIRP_FAIL_ error code see 
 *   psirpd_types.h
 */
psirp_error_t 
psirpd_rzv_rcv_subscribe_metadata(pkt_ctx_t *pkt_ctx) {
    
    //psirpd_hdrs_rzvhdr_t  *rzv_hdr;
    psirpd_hdrs_metadata_t *md_hdr;

    ENTER();

    PSIRP_ETL(pkt_ctx == NULL, PSIRP_FAIL_NULL_POINTER, 
              PSIRP_DBG_ERR, "pkt_ctx == NULL");

    md_hdr = pkt_ctx->md_hdr;
    PSIRP_ETL(md_hdr == NULL, PSIRP_FAIL_NULL_POINTER, 
              PSIRP_DBG_ERR, "pkt_ctx->md_hdr == NULL");

#if 0
    rzv_hdr = pkt_ctx->rzv_hdr;
    PSIRP_ETL(rzv_hdr == NULL, PSIRP_FAIL_NULL_POINTER, 
              PSIRP_DBG_ERR, "pkt_ctx->rzv_hdr == NULL");
              PSIRP_DBG_ERR, "pkt_ctx->rzv_hdr == NULL");
    /* We should not ignore the RZV header. For example, we could take
     * the SId and RId and publish the rest of the packet with those
     * Ids and expect that some entity will process the packet
     * further. */
#endif

    PSIRP_EF(psirpd_ipc_net_handle_metadata(md_hdr));

    RETURN PSIRP_OK;
}

/** 
 * Handler that is called for outgoing data chunk or 
 * metadata subscription packet.
 * @param pkt_ctx pointer to the packet context
 * @return On sucess: PSIRP_OK <br>On failure: PSIRP_FAIL_ error code see 
 *   psirpd_types.h
 */
psirp_error_t 
psirpd_rzv_snd_subscribe(pkt_ctx_t *pkt_ctx) { /* XXX: ? */

    ENTER();
#if 0 /* done in psirpd_packet_out_* */
    psirpd_hdrs_rzvhdr_t   *rzv_hdr;
    psirpd_hdrs_metadata_t *md_hdr;
    int i;

    ENTER();

    PSIRP_ETL(pkt_ctx == NULL, PSIRP_FAIL_NULL_POINTER, 
              PSIRP_DBG_ERR, "pkt_ctx == NULL");

    rzv_hdr = pkt_ctx->rzv_hdr;
    PSIRP_ETL(rzv_hdr == NULL, PSIRP_FAIL_NULL_POINTER, 
              PSIRP_DBG_ERR, "pkt_ctx->rzv_hdr == NULL");
 
    /* TODO: RID and SID are got during the network attachment.
     *  Now we just add a random RID and SID. */
    for (i=0; i < sizeof(psirp_id_t); i++){
        rzv_hdr->rid.rid[i]=(u_int8_t)random();
    }
    for (i=0; i < sizeof(psirp_sid_t); i++){
        rzv_hdr->sid.sid[i]=(u_int8_t)random();
    }


    if (pkt_ctx->data_map) {
        md_hdr = pkt_ctx->md_hdr;
        PSIRP_ETL(md_hdr == NULL, PSIRP_FAIL_NULL_POINTER, 
                  PSIRP_DBG_ERR, "pkt_ctx->md_hdr == NULL");
        PSIRP_EF(psirpd_bmap_export(pkt_ctx->data_map, md_hdr->data_map, 
                                    PSIRP_DATAMAP_SIZE));
    }
#endif

    RETURN PSIRP_OK;
}

/** 
 * Handler that is called for outgoing metadata packet
 * @param pkt_ctx pointer to the packet context
 * @return On sucess: PSIRP_OK <br>On failure: PSIRP_FAIL_ error code see 
 *   psirpd_types.h
 */
psirp_error_t 
psirpd_rzv_snd_metadata(pkt_ctx_t *pkt_ctx) { /* XXX: ? */

    ENTER();
#if 0

    psirpd_hdrs_metadata_t *md_hdr;

    ENTER();

    PSIRP_ETL(pkt_ctx == NULL, PSIRP_FAIL_NULL_POINTER, 
              PSIRP_DBG_ERR, "pkt_ctx == NULL");

    md_hdr = pkt_ctx->md_hdr;
    PSIRP_ETL(md_hdr == NULL, PSIRP_FAIL_NULL_POINTER, 
              PSIRP_DBG_ERR, "pkt_ctx->md_hdr == NULL");

    /* Copy the publication len to the metadata */
    md_hdr->len = pkt_ctx->pub->md->len;
    PSIRP_ETL(md_hdr->len == 0 || md_hdr->len > PSIRP_MAX_PUB_SIZE, 
              PSIRP_FAIL_BUF_LEN,
              PSIRP_DBG_ERR, "md_hdr->len == %d", md_hdr->len);

    /* copy the max sequence number to the metadata */
    md_hdr->max_seqnum = pkt_ctx->pub->md->max_seqnum;
    PSIRP_ETL(md_hdr->max_seqnum > PSIRP_DATAMAP_SIZE, 
              PSIRP_FAIL_BUF_LEN, 
              PSIRP_DBG_ERR, "md_hdr->max_seqnum == %d", md_hdr->max_seqnum);

    /* Copy the data bitmap to the metadata */
    PSIRP_ETL(pkt_ctx->data_map == NULL, PSIRP_FAIL_NULL_POINTER, 
              PSIRP_DBG_ERR, "pkt_ctx->sub_map == NULL");
    PSIRP_EF(psirpd_bmap_export(pkt_ctx->data_map, md_hdr->data_map, 
                                PSIRP_DATAMAP_SIZE));
#endif
    
    RETURN PSIRP_OK;
}

